#' Sequential backward search method
#'
#' @description This function sequentially eliminates one variant within the optimal region located by ZFA at each iteration until obtaining a subset of variants that are enriched for association with the phenotype.
#'
#' @param data a data frame or numeric matrix. Genotypes should be coded as 0,1 or 2.
#' @param y a numeric vector with two levels. Phenotype values are coded as 0 or 1.
#' @param zfaResult results of ZFA optimization, the output of "zfa" and "fuzzy.zfa" functions.
#' @param test a character to choose the rare variant method that combines with the Zoom- Focus Algorithm.If test = "SKAT", the SKAT of variance component test is applied. If test = "SKATO", the SKAT-O of combination method test is applied. If test = "burden", the weighted burden test is applied. If test = "wtest", the W-test of burden test category is applied.
#' @param fuzzy a logical value indicating whether zfaResult is the result of "fuzzy.zfa" function. Default = FALSE.
#'
#' @details This function further refines the optimal region located by ZFA.
#'
#' @return The function returns a list with two components:
#'
#' \item{sbs.result}{a data frame of results aggregated over the sequential backward search process.}
#'
#' \item{sbs.best}{a list including the index of removed and selected variants, and the p-value of the selected variants. Note that the p-value has already been adjusted by Bonferroni correction.}
#'
#' @examples
#'
#' data(zfa.example) 
#' attach(zfa.example)
#' 
#' ## fast-zoom with wtest, all zoomed regions passed to focusing step, 
#' ## and output all results
#' zfa.result<-zfa(X,y,bin = 32,fast.path = TRUE, 
#'             filter.pval=NULL,output.pval=NULL,test = "wtest")
#'
#' ## further refines the optimal region located by ZFA.
#' zfa.sbs <- sbs.search(X,y,zfa.result,"wtest",fuzzy = FALSE)
#'
#' @export

sbs.search <- function(data,y,zfaResult,test,fuzzy=FALSE){
  if(!fuzzy){
    zfa_region <- c(zfaResult$result[1,1]:zfaResult$result[1,2])
    zfa_steps <- zfaResult$result[1,5]
  }else{
    zfa_region <- c(zfaResult[,1]:zfaResult[,2])
    zfa_steps <- zfaResult[,6]
  }
  obj <- list()
  if(test!="test"){
    obj <- SKAT::SKAT_Null_Model(y ~ 1, out_type="D")
  }
  sbs_results <- data.frame(sbs(data,y,obj,test,zfa_region))
  colnames(sbs_results) <- c("remove.index","p.val")
  sbs_results[1,1] <- NA
  sbs_results[,2] <- sbs_results[,2]*(zfa_steps+length(zfa_region)*(length(zfa_region)+1)/2-1)# p.val adjust
  sbs_results[,2] <- ifelse(sbs_results[,2]>1,1,sbs_results[,2])
  min_indx <- which(sbs_results[,2]==min(sbs_results[,2]))
  if(length(min_indx)==1 && min_indx[1]==1){
    removed <- NA
  }else{
    removed <- sbs_results[2:min_indx[length(min_indx)],1]
  }
  selected <- setdiff(zfa_region,removed)
  sbs_result <- list(sbs.results = sbs_results, sbs.best = list(removed = sbs_results[2:which.min(sbs_results[,2]),1], selected = selected, p.val = min(sbs_results[,2])))
  sbs_result
}
