#' Test of single or collapsed variants
#'
#' @description This function performs either single-variant association tests for desired method or region-based rare variant test of collapsed variants. Four alternative methods are supported in the current version.
#'
#' @param data a data frame or numeric matrix, usually a gene. Genotypes should be coded as 0,1 or 2.
#' @param y a numeric vector with two levels. Phenotype values are coded as 0 or 1.
#' @param collapse a logical value indicating whether or not to collapse all the variants for association testing. Default = \code{FALSE} for single variant tests, otherwise \code{TRUE} represents collapsed association test.
#' @param test a character to choose the rare variant method that combines with the Zoom-Focus Algorithm.If test = "SKAT", the SKAT of variance component test is applied. If test = "SKATO", the SKAT-O of combination method test is applied. If test = "burden", the weighted burden test is applied. If test = "wtest", the W-test of burden test category is applied.
#' @return The output depends on \code{collapse} setting:
#'
#' When \code{collapse=FALSE}, it returns a data frame of following components:
#'
#' \item{pos}{Array position/name of variant}
#'
#' \item{odds.ratio}{Odds ratio of variant}
#'
#' \item{p.value}{P-value of single variant association test}
#'
#' When \code{collapse=TRUE}, it returns a table of following elements:
#'
#' \item{p.value}{P-value of gene-based rare variant test}
#'
#' \item{odds.ratio}{Regional odds ratio of collapsed variants within genomic region, see \code{\link{odds.ratio}} for more details}
#'
#' \item{n.variant}{Total number of variants within test region}
#'
#' \item{test}{Type of selected rare variant method}
#'
#' @references
#'
#' M. H. Wang., R. Sun., et al. (2016) A fast and powerful W-test for pairwise epistasis testing. Nucleic Acids Research.doi:10.1093/nar/gkw347.
#'
#' R. Sun., H. Weng., et al. (2016) A W-test collapsing method for rare-variant association testing in exome sequencing data. Genetic Epidemiology, 40(7): 591-596.
#'
#' Wu, M. C., Lee, S., et al. (2011) Rare Variant Association Testing for Sequencing Data Using the Sequence Kernel Association Test (SKAT). American Journal of Human Genetics, 89, 82-93.
#'
#' Lee, S., Emond, M.J., et al. (2012) Optimal unified approach for rare variant association testing with application to small sample case-control whole-exome sequencing studies. American Journal of Human Genetics, 91, 224-237.
#'
#' Lee, S., Fuchsberger, C., et al. (2015) An efficient resampling method for calibrating single and gene-based rare variant association analysis in case-control studies. Biostatistics, kxv033.
#'
#' @examples
#' data(fuzzy.zfa.example)
#' attach(fuzzy.zfa.example)
#'
#' # Single variant tests by wtest
#' result.single.variants<-test.variants(Z,y,collapse=FALSE,test = "wtest")
#'
#' # Collapsed association test by SKAT
#' result.collapsed.variants<-test.variants(Z,y,collapse=TRUE,test = "SKAT")
#'
#' @export
#' @seealso \code{\link{fuzzy.zfa}}, \code{\link{odds.ratio}}, \code{\link{zfa.plot}}
#' @importFrom stats pchisq


test.variants<-function(data,y,collapse=FALSE,test=c("SKAT","SKATO","burden","wtest")) {
  if (collapse==FALSE) {
    if (test=="wtest") {
      pvalue<-unname(apply(data,2,function(x) wtest(as.matrix(x),y)))
    } else {
      obj<-SKAT::SKAT_Null_Model(y ~ 1, out_type="D")
      pvalue<-NULL
      for (i in 1:ncol(data)) {
        pvalue[i]<-(SKAT::SKATBinary(as.matrix(data[,i]),obj)$p.value)
      }
    }
    OR<-odds.ratio(data,y,collapse=FALSE)
    result<-data.frame(OR,p.value=pvalue)
  } else {
    if (test=="wtest") {
      pvalue<-wtest(data,y)
    } else if (test=="SKAT") {
      obj<-SKAT::SKAT_Null_Model(y ~ 1, out_type="D")
      pvalue<-SKAT::SKATBinary(as.matrix(data),obj)$p.value
    } else if (test=="SKATO") {
      obj<-SKAT::SKAT_Null_Model(y ~ 1, out_type="D")
      pvalue<-SKAT::SKATBinary(as.matrix(data),obj,method="SKATO")$p.value
    } else if (test=="burden") {
      obj<-SKAT::SKAT_Null_Model(y ~ 1, out_type="D")
      pvalue<-SKAT::SKATBinary(as.matrix(data),obj,method="Burden")$p.value
    }
    OR<-odds.ratio(data,y,collapse=TRUE)
    result<-data.frame(p.value=pvalue,odds.ratio=OR,n.variant=ncol(data),test=test)
  }
  return(result)
}


